from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import requests
import logging
import json

from typing import Any

from rasa.nlu.components import Component
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.training_data.message import Message

logger = logging.getLogger(__name__)


class BingSpellCheck(Component):
    name = 'BingSpellCheck'

    defaults = {
        'key': '',
        'min_score': 0.8,
        'language': 'en-US',
        'enabled': True,
    }

    def __init__(self, component_config=None):
        # type: (RasaNLUModelConfig) -> None
        super(BingSpellCheck, self).__init__(component_config)

        self.url = 'https://api.cognitive.microsoft.com/bing/v7.0/spellcheck/'
        self.header = {
            'Ocp-Apim-Subscription-Key': self.component_config['key'],
            'setLang': self.component_config['language'][:2],
        }

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None

        text = message.text
        response = self._response(text)
        tokens = response.get('flaggedTokens', [])

        replacements = self._get_replacements(tokens)

        if self.component_config.get('enabled', True):
            message.text = self._replace(text, replacements)

    def _response(self, text):
        # type (str) -> dict
        default_response = {
          '_type': 'SpellCheck',
          'flaggedTokens': []
        }

        try:
            response = requests.post(self.url, data=self._payload(text), headers=self.header)
            if response.status_code == 200:
                return json.loads(response.text)
            else:
                logger.error("Failed to get a proper response from spell check "
                             "server {}. Status Code: {}. Response: {}"
                             "".format(self.url, response.status_code, response.text))

                return default_response
        except requests.exceptions.ConnectionError as e:
            logger.error("Failed to connect to the spell check http server. "
                         "More information at "
                         "https://azure.microsoft.com/en-us/services/cognitive-services/spell-check/"
                         "Error: {}".format(e))
            return default_response

    def _payload(self, text):
        return {
            'text': text,
            'mode': 'spell',
            'mkt': self.component_config['language'],
        }

    @staticmethod
    def _replace(text, replacements):
        replacements.sort(key=lambda x: x['offset'])
        start_next = len(text)

        while len(replacements):
            flagged_token = replacements.pop()

            offset = flagged_token['offset']
            token = flagged_token['token']
            replacement = flagged_token['replacement']

            if offset + len(token) <= start_next:
                text = text[0:offset] + replacement + text[offset + len(token):]
                start_next = offset

        return text

    def _get_replacements(self, tokens):
        replacements = []
        for flagged_token in tokens:

            offset = flagged_token['offset']
            token = flagged_token['token']
            suggestions = sorted(flagged_token['suggestions'], key=lambda x: x['score'], reverse=True)

            if len(suggestions) and suggestions[0]['score'] > self.component_config['min_score']:
                replacements.append({
                    'offset': offset,
                    'token': token,
                    'replacement': suggestions[0]['suggestion']
                })

        return replacements


import time

import logging
import os
import requests
import simplejson
from typing import Any, List, Optional, Text, Dict

from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.extractors import EntityExtractor
from rasa.nlu.model import Metadata
from rasa.nlu.training_data import Message

logger = logging.getLogger(__name__)


def extract_value(match):
    if match["value"].get("type") == "interval":
        value = {"to": match["value"].get("to", {}).get("value"),
                 "from": match["value"].get("from", {}).get("value")}
    else:
        value = match["value"].get("value")

    return value


def filter_irrelevant_matches(matches, requested_dimensions):
    """Only return dimensions the user configured"""

    if requested_dimensions:
        return [match
                for match in matches
                if match["dim"] in requested_dimensions]
    else:
        return matches


def convert_duckling_format_to_rasa(matches):
    extracted = []

    for match in matches:
        value = extract_value(match)
        entity = {"start": match["start"],
                  "end": match["end"],
                  "text": match.get("body", match.get("text", None)),
                  "value": value,
                  "confidence": 1.0,
                  "additional_info": match["value"],
                  "entity": match["dim"]}

        extracted.append(entity)

    return extracted


class DucklingHTTPExtractor(EntityExtractor):
    """Searches for structured entites, e.g. dates, using a duckling server."""

    name = "DucklingHTTPExtractor"

    provides = ["entities"]

    defaults = {
        # by default all dimensions recognized by duckling are returned
        # dimensions can be configured to contain an array of strings
        # with the names of the dimensions to filter for
        "dimensions": None,

        # http url of the running duckling server
        "url": None,

        # locale - if not set, we will use the language of the model
        "locale": None,

        # timezone like Europe/Berlin
        # if not set the default timezone of Duckling is going to be used
        "timezone": None
    }

    def __init__(self,
                 component_config: Text = None,
                 language: Optional[List[Text]] = None) -> None:

        super(DucklingHTTPExtractor, self).__init__(component_config)
        self.language = language

    @classmethod
    def create(
        cls, component_config: Dict[Text, Any], config: RasaNLUModelConfig
    ) -> 'DucklingHTTPExtractor':
        return cls(component_config,
                   config.language)

    def _locale(self):
        if not self.component_config.get("locale"):
            # this is king of a quick fix to generate a proper locale
            # works most of the time
            locale_fix = "{}_{}".format(self.language, self.language.upper())
            self.component_config["locale"] = locale_fix
        return self.component_config.get("locale")

    def _url(self):
        """Return url of the duckling service. Environment var will override."""
        if os.environ.get("RASA_DUCKLING_HTTP_URL"):
            return os.environ["RASA_DUCKLING_HTTP_URL"]

        return self.component_config.get("url")

    def _payload(self, text, reference_time, timezone):
        return {
            "text": text,
            "locale": self._locale(),
            "tz": timezone,
            "reftime": reference_time
        }

    def _duckling_parse(self, text, reference_time, timezone):
        """Sends the request to the duckling server and parses the result."""

        try:
            payload = self._payload(text, reference_time, timezone)
            headers = {"Content-Type": "application/x-www-form-urlencoded; "
                                       "charset=UTF-8"}
            response = requests.post(self._url() + "/parse",
                                     data=payload,
                                     headers=headers)
            if response.status_code == 200:
                return simplejson.loads(response.text)
            else:
                logger.error("Failed to get a proper response from remote "
                             "duckling. Status Code: {}. Response: {}"
                             "".format(response.status_code, response.text))
                return []
        except requests.exceptions.ConnectionError as e:
            logger.error("Failed to connect to duckling http server. Make sure "
                         "the duckling server is running and the proper host "
                         "and port are set in the configuration. More "
                         "information on how to run the server can be found on "
                         "github: "
                         "https://github.com/facebook/duckling#quickstart "
                         "Error: {}".format(e))
            return []

    @staticmethod
    def _timezone_from_config_or_request(component_config, timezone):
        if timezone is not None:
            return timezone
        return component_config.get("timezone")

    @staticmethod
    def _reference_time_from_message_or_request(message, reference_time):
        if message.time is not None:
            try:
                return int(message.time) * 1000
            except ValueError as e:
                logging.warning("Could not parse timestamp {}. Instead "
                                "current UTC time will be passed to "
                                "duckling. Error: {}".format(message.time, e))
        # fallbacks to current time, multiplied by 1000 because duckling
        # requires the reftime in milliseconds
        elif reference_time is not None:
            try:
                return int(reference_time)
            except ValueError as e:
                logging.warning("Could not parse timestamp {}. Instead "
                                "current UTC time will be passed to "
                                "duckling. Error: {}".format(reference_time, e))
        return int(time.time()) * 1000

    def process(self, message: Message, **kwargs: Any) -> None:

        if self._url() is not None:
            params = kwargs
            timezone = self._timezone_from_config_or_request(
                self.component_config, params.get("timezone", None))
            reference_time = self._reference_time_from_message_or_request(
                message, params.get("reference_time", None))
            matches = self._duckling_parse(message.text, reference_time,
                                           timezone)
            dimensions = self.component_config["dimensions"]
            relevant_matches = filter_irrelevant_matches(matches, dimensions)
            extracted = convert_duckling_format_to_rasa(relevant_matches)
        else:
            extracted = []
            logger.warning("Duckling HTTP component in pipeline, but no "
                           "`url` configuration in the config "
                           "file nor is `RASA_DUCKLING_HTTP_URL` "
                           "set as an environment variable.")

        extracted = self.add_extractor_name(extracted)
        message.set("entities",
                    message.get("entities", []) + extracted,
                    add_to_output=True)

    @classmethod
    def load(cls,
             component_meta: Dict[Text, Any],
             model_dir: Text = None,
             model_metadata: Metadata = None,
             cached_component: Optional['DucklingHTTPExtractor'] = None,
             **kwargs: Any
             ) -> 'DucklingHTTPExtractor':
        return cls(component_meta, model_metadata.get("language"))



import os
import warnings
import rasa

from typing import Any, Text, Dict, Optional

from rasa.nlu.components import Component
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.training_data import Message, TrainingData
from rasa.nlu.model import Metadata

from fuzzy_matcher import process

class Gazette(Component):
    name = "Gazette"

    provides = ["entities"]

    defaults = {
        "max_num_suggestions": 5,
        "entities": [],
    }

    def __init__(self,
        component_config: Text = None,
        gazette: Optional[Dict] = None) -> None:

        super(Gazette, self).__init__(component_config)
        self.gazette = gazette if gazette else {}
        if gazette: self._load_config()
        self.limit = self.component_config.get("max_num_suggestions")
        self.entities = self.component_config.get("entities", [])

    def process(self, message: Message, **kwargs: Any) -> None:

        entities = message.get("entities", [])
        new_entities = []

        for entity in entities:
            config = self._find_entity(entity, self.entities)
            if config is None or not isinstance(entity["value"], str):
                new_entities.append(entity)
                continue

            matches = process.extract(
                entity["value"],
                self.gazette.get(entity["entity"], []),
                limit=self.limit,
                scorer=config["mode"]
            )
            primary, score = matches[0] if len(matches) else (None, None)

            if primary is not None and score > config["min_score"]:
                entity["value"] = primary
                entity["gazette_matches"] = [{"value": value, "score": num} for value, num in matches]
                new_entities.append(entity)

        message.set("entities", new_entities)

    def train(
        self, training_data: TrainingData, cfg: RasaNLUModelConfig, **kwargs: Any
    ) -> None:
        self.gazette = self._load_gazette_list(training_data.gazette)

    @classmethod
    def load(cls,
        component_meta: Dict[Text, Any],
        model_dir: Text = None,
        model_metadata: Metadata = None,
        cached_component: Optional['Gazette'] = None,
        **kwargs: Any
    ) -> 'Gazette':
        td = rasa.utils.io.read_json_file(os.path.join(model_dir, "training_data.json"))
        if "gazette" in td["rasa_nlu_data"]:
            gazette = cls._load_gazette_list(td["rasa_nlu_data"]["gazette"])
        else:
            gazette = None
            warnings.warn("Could not find Gazette in persisted training data file.")

        return Gazette(component_meta, gazette)

    @staticmethod
    def _load_gazette_list(gazette: Optional[Dict]) -> None:
        gazette_dict = {}
        for item in gazette:
            name = item["value"]
            table = item["gazette"]
            gazette_dict[name] = table
        return gazette_dict

    @staticmethod
    def _find_entity(entity, entities):
        for rep in entities:
            if entity["entity"] == rep["name"]:
                return rep
        return None

    def _load_config(self):
        entities = []
        for rep in self.component_config.get("entities", []):
            assert "name" in rep, "Must provide the entity name for the gazette entity configuration: {}".format(rep)
            assert rep["name"] in self.gazette, "Could not find entity name {0} in gazette {1}".format(rep["name"], self.gazette)

            supported_properties = ["mode", "min_score"]
            defaults = ["ratio", 80]
            types = [str, int]

            new_element = {"name": rep["name"]}
            for prop, default, t in zip(supported_properties, defaults, types):
                if prop not in rep:
                    new_element[prop] = default
                else:
                    new_element[prop] = t(rep[prop])

            entities.append(new_element)

        self.component_config["entities"] = entities


from rasa.core.events import Event
from rasa.core.actions.action import Action, ActionUtterTemplate, create_bot_utterance

from typing import Any, List, Text, Dict, Optional
import logging
import copy

from requests.auth import HTTPBasicAuth

logging.basicConfig(level="WARN")
logger = logging.getLogger()

class ActionBotfrontMapping(Action):
    def name(self):
        return 'action_botfront_mapping'

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
    ) -> List[Event]:
        """Append 'utter_' to intent name and generates from that template"""

        events = []
        response_name = 'utter_' + tracker.latest_message.intent["name"]
        events += [create_bot_utterance(m) for m in await nlg.generate(response_name, tracker, output_channel.name(),
                   language=tracker.latest_message.metadata["language"])]

        return events



import logging
import json
import os
import re
import copy
from typing import Any, List, Text, Dict

import rasa.utils.io

from rasa.core.actions.action import (
    ACTION_BACK_NAME,
    ACTION_LISTEN_NAME,
    ACTION_RESTART_NAME,
)
from rasa.core.constants import USER_INTENT_BACK, USER_INTENT_RESTART
from rasa.core.domain import Domain
from rasa.core.events import ActionExecuted
from rasa.core.policies.policy import Policy
from rasa.core.trackers import DialogueStateTracker

logger = logging.getLogger(__name__)


class BotfrontMappingPolicy(Policy):
    """Policy which maps intents directly to actions.

    Like RasaHQ's, except it looks for regex in the intent
    name, and always triggers the same premade action,
    meaning the model doesn't have to be retrained every
    time a new mapped intent is defined."""
    
    defaults = {
        'triggers': [{'trigger': r'^map\..+', 'action': 'action_botfront_mapping'}]
    }

    def __init__(self, priority: int = 999, **kwargs: Any) -> None:
        super(BotfrontMappingPolicy, self).__init__(priority=priority)
        self._load_params(**kwargs)

    def _load_params(self, **kwargs: Dict[Text, Any]) -> None:
        config = copy.deepcopy(self.defaults)
        config.update(kwargs)
        triggers = config.pop('triggers')
        if isinstance(triggers, list):
            self.triggers = {trigger['trigger']: trigger['action'] for trigger in triggers}
        else:
            self.triggers = triggers

    def train(
        self,
        training_trackers: List[DialogueStateTracker],
        domain: Domain,
        **kwargs: Any
    ) -> None:
        """Does nothing. This policy is deterministic."""

        pass

    def predict_action_probabilities(
        self, tracker: DialogueStateTracker, domain: Domain
    ) -> List[float]:
        logger.debug('Triggers: ' + ', '.join(self.triggers.keys()))
        """Predicts the assigned action.

        If the current intent is assigned to an action that action will be
        predicted with the highest probability of all policies. If it is not
        the policy will predict zero for every action."""

        prediction = [0.0] * domain.num_actions
        intent = tracker.latest_message.intent.get("name")
        action = None
        if isinstance(intent, str):
            for trigger in self.triggers:
                match = re.search(trigger, intent)
                if match: action = self.triggers[trigger]
        if tracker.latest_action_name == ACTION_LISTEN_NAME:
            if action:
                idx = domain.index_for_action(action)
                if idx is None:
                    logger.warning(
                        "{} is not defined.".format(action)
                    )
                else:
                    prediction[idx] = 1
            elif intent == USER_INTENT_RESTART:
                idx = domain.index_for_action(ACTION_RESTART_NAME)
                prediction[idx] = 1
            elif intent == USER_INTENT_BACK:
                idx = domain.index_for_action(ACTION_BACK_NAME)
                prediction[idx] = 1

            if any(prediction):
                logger.debug(
                    "The predicted intent '{}' is being "
                    " handled by BotfrontMappingPolicy."
                    "".format(intent)
                )
        elif tracker.latest_action_name == action and action is not None:
            latest_action = tracker.get_last_event_for(ActionExecuted)
            assert latest_action.action_name == action

            if latest_action.policy == type(self).__name__:
                # this ensures that we only predict listen, if we predicted
                # the mapped action
                logger.debug(
                    "BotfrontMappingPolicy has just been triggered, "
                    "so now returning to action_listen. "
                )

                idx = domain.index_for_action(ACTION_LISTEN_NAME)
                prediction[idx] = 1
        else:
            logger.debug(
                "Predicted intent is not handled by BotfrontMappingPolicy."
            )
        return prediction

    def persist(self, path: Text) -> None:
        """Persists priority and trigger regex"""

        config_file = os.path.join(path, "botfront_mapping_policy.json")
        meta = {
            "priority": self.priority,
            "triggers": self.triggers
        }
        rasa.utils.io.create_directory_for_file(config_file)
        rasa.utils.io.dump_obj_as_json_to_file(config_file, meta)

    @classmethod
    def load(cls, path: Text) -> "BotfrontMappingPolicy":
        """Returns the class with the configured priority."""

        meta = {}
        if os.path.exists(path):
            meta_path = os.path.join(path, "botfront_mapping_policy.json")
            if os.path.isfile(meta_path):
                meta = json.loads(rasa.utils.io.read_file(meta_path))

        return cls(**meta)



import logging
from rasa.core.events import UserUttered
from rasa_addons.core.tracker_stores.botfront import BotfrontTrackerStore
from rasa_addons.core.tracker_stores.askhub import AskHubAnalytics
from rasa_addons.core.tracker_stores.chatbase import ChatbaseAnalytics
from rasa_addons.utils import get_latest_parse_data_language

logger = logging.getLogger(__name__)


class AnalyticsTrackerStore(BotfrontTrackerStore):
    def __init__(self, domain, url, **kwargs):
        super(AnalyticsTrackerStore, self).__init__(domain, url, **kwargs)
        self.chatbase_api_key = kwargs.get('chatbase_api_key')
        self.chatbase_version = str(kwargs.get('chatbase_version', 0))
        self.askhub_api_key = kwargs.get('askhub_api_key')
        self.askhub_response_format = kwargs.get('askhub_response_format', 'botfront')
        self.latest_response_name = kwargs.get('latest_response_name', 'latest_response_name')
        self.special_action_responses = kwargs.get('special_actions_responses', [])

    def save(self, tracker):
        all_events = super(AnalyticsTrackerStore, self).save(tracker)
        platform = self.get_latest_input_channel(tracker)
        if self.is_end_of_turn(all_events[-1]):
            latest_turn = self.get_latest_turn(all_events)
            if latest_turn:
                if self.askhub_api_key:
                    AskHubAnalytics().save(latest_turn, tracker, all_events, platform, self)
                if self.chatbase_api_key:
                    ChatbaseAnalytics().save(latest_turn, tracker, all_events, platform, self)

    def get_latest_turn(self, all_events):
        events = reversed(all_events)
        latest_turn = {
            'bot_events': [],
            'response_names': [],
        }
        try:
            while True:
                event = next(events)
                # print (evt)
                if event['event'] == 'bot':
                    latest_turn['bot_events'].insert(0, event)
                # Get response name from FAQ action
                if event['event'] == 'slot' and event['name'] == 'latest_response_name':
                    latest_turn['response_names'].insert(0, event['value'])
                # Get response name from utter action
                if event['event'] == 'action' and event['name'].startswith('utter'):
                    latest_turn['response_names'].insert(0, event['name'])
                # Get special action response names
                special_actions = list(map(lambda a: a['action_name'], self.special_action_responses))
                if event['event'] == 'action' and event['name'] in special_actions:
                    response_name = next((x for x in self.special_action_responses
                                        if x['action_name'] == event['name']))['response_name']
                    latest_turn['response_names'].insert(0, response_name)
                if event['event'] == 'user':
                    latest_turn['user_event'] = event
                    break

            latest_turn['parse_data'] = AnalyticsTrackerStore.get_parse_data(latest_turn)
            latest_turn['language'] = get_latest_parse_data_language(all_events)
            intent = AnalyticsTrackerStore.get_latest_intent(latest_turn['parse_data'])
            if intent:
                latest_turn['intent'] = intent
        except StopIteration:
            return None
        return latest_turn

    def is_latest_bot_response(self, event):
        return event['event'] == 'slot' and event['name'] == self.latest_response_name

    def is_fallback(self, event):
        return event['event'] == 'action' and event['name'] == self.fallback_action_name

    @staticmethod
    def is_end_of_turn(event):
        return event['event'] == 'action' and event['name'] == 'action_listen'

    @staticmethod
    def get_parse_data(latest_turn):
        parse_data = latest_turn['user_event']['parse_data']
        if 'original_data' in parse_data:
            parse_data = parse_data['original_data']
        return parse_data

    @staticmethod
    def get_latest_intent(parse_data):
        return parse_data['intent']['name'] if 'intent' in parse_data and 'name' in parse_data['intent'] else None

    @staticmethod
    def get_latest_input_channel(tracker):
        for event in reversed(tracker.events):
            if isinstance(event, UserUttered) and event.input_channel is not None:
                return event.input_channel



import logging
import jsonpickle
import requests
import time
from threading import Thread

from rasa.core.tracker_store import TrackerStore
from rasa.core.trackers import DialogueStateTracker, EventVerbosity

logger = logging.getLogger(__name__)

jsonpickle.set_preferred_backend('json')
jsonpickle.set_encoder_options('json', ensure_ascii=False)


class UnsynchronizedTracker:
    # Want these objects to behave just like tracker dict
    def __init__(self, tracker=None):
        self.tracker = tracker

    def __getitem__(self, i):
        return self.tracker[i]

    def __setitem__(self, i, value):
        self.tracker[i] = value

    def is_initialized(self):
        return self.tracker is not None


def _start_sweeper(tracker_store, break_time):
    while True:
        try:
            tracker_store.sweep()
        finally:
            time.sleep(break_time)


class BotfrontTrackerStore(TrackerStore):
    def __init__(self, domain, url, **kwargs):
        self.front_project_id = kwargs.get('project_id')
        self.front_url = url
        self.tracker_persist_time = kwargs.get('tracker_persist_time', 3600)
        self.max_events = kwargs.get('max_events', 100)

        self.trackers = {}
        self.sweeper = Thread(target=_start_sweeper, args=(self, 30))
        self.sweeper.setDaemon(True)
        self.sweeper.start()

        super(BotfrontTrackerStore, self).__init__(domain)

    def _fetch_tracker(self, sender_id):
        if not sender_id in self.trackers:
            tracker = self._get(sender_id)
            self.trackers[sender_id] = tracker

        return self.trackers[sender_id]

    def _url(self, sender_id, append):
        return '{url}/project/{project_id}/conversations/{sender_id}/{append}'.format(url=self.front_url,
                                                                                      project_id=self.front_project_id,
                                                                                      sender_id=sender_id,
                                                                                      append=append
                                                                                      )

    def _get(self, sender_id):
        logger.debug('{} -> GET request for tracker'.format(sender_id))

        try:
            response = requests.get(self._url(sender_id, self.max_events), timeout=1)
            if response.status_code == 200:
                return response.json()  # Tracker object or None if none in database

            message = response.json() if response.status_code == 400 else response.reason
            logger.warning(
                'Error: get request to botfront-api failed with status code {}: {}'.format(response.status_code,
                                                                                           message))
        except Exception as e:
            logger.error(
                'Error: get request to botfront-api failed with error: {}'.format(e))

        return UnsynchronizedTracker()  # Return an object to represent synchronization failure with database

    def _post(self, sender_id, method, payload):
        url = self._url(sender_id, method)

        try:
            response = requests.post(url, json=payload, timeout=1)
            if response.status_code != 200:
                message = response.json() if response.status_code == 400 else response.reason
                logger.warning(
                    "Error: post request to botfront-api failed with status code {}: {}".format(response.status_code,
                                                                                                message))
        except Exception as e:
            logger.error(
                'Error: post request to botfront-api failed with error: {}'.format(e))

    @staticmethod
    def _is_none(tracker):
        return tracker is None or (isinstance(tracker, UnsynchronizedTracker) and not tracker.is_initialized())

    def save(self, canonical_tracker):

        # Fetch here just in case retrieve wasn't called first
        sender_id = canonical_tracker.sender_id
        tracker = self._fetch_tracker(sender_id)
        serialized_tracker = self._serialize_tracker_to_dict(canonical_tracker)

        if self._is_none(tracker):
            # No document in the collection, so insert one

            if not isinstance(tracker, UnsynchronizedTracker):
                logger.debug('{} -> INSERTING TRACKER: {}'.format(sender_id, serialized_tracker))
                self._post(sender_id, 'insert', serialized_tracker)
                self.trackers[sender_id] = serialized_tracker
            else:
                logger.debug('{} -> UNSYNCHRONIZED TRACKER UPDATE'.format(sender_id))
                self.trackers[sender_id] = UnsynchronizedTracker(serialized_tracker)

            return serialized_tracker['events']

        # Insert only the new examples
        last_timestamp = (next(reversed(tracker['events']), {})).get('timestamp', None)
        new_events = list(filter(lambda x: x['timestamp'] > last_timestamp, serialized_tracker['events']))

        if not isinstance(tracker, UnsynchronizedTracker):
            tracker_shallow_copy = {key: val for key, val in serialized_tracker.items()}
            tracker_shallow_copy['events'] = new_events

            logger.debug('{} -> UPDATING TRACKER: {}'.format(sender_id, tracker_shallow_copy))
            self._post(sender_id, 'update', tracker_shallow_copy)
            self.trackers[sender_id] = serialized_tracker
        else:
            logger.debug('{} -> UNSYNCHRONIZED TRACKER UPDATE'.format(sender_id))
            self.trackers[sender_id] = UnsynchronizedTracker(serialized_tracker)

        return serialized_tracker['events']

    def _init_tracker(self, sender_id, tracker):
        if self.domain:
            return DialogueStateTracker.from_dict(sender_id,
                                                  tracker["events"],
                                                  self.domain.slots)

        else:
            logger.warning("Can't recreate tracker from mongo storage "
                           "because no domain is set. Returning `None` "
                           "instead.")
            return None

    def retrieve(self, sender_id):
        tracker = self._fetch_tracker(sender_id)

        if not self._is_none(tracker):
            return self._init_tracker(sender_id, tracker)

        return None

    def sweep(self):
        # Iterate over list of keys to prevent runtime errors when deleting elements
        for key in list(self.trackers.keys()):
            tracker = self.trackers[key]
            if not self._is_none(tracker) and tracker['latest_event_time'] < time.time() - self.tracker_persist_time:
                logger.debug('SWEEPER: Removing tracker for user {}'.format(key))
                del self.trackers[key]

    @staticmethod
    def _serialize_tracker_to_dict(canonical_tracker):
        return canonical_tracker.current_state(EventVerbosity.ALL)





import asyncio
import rasa
import logging
import json
import inspect
from rasa.core.channels.channel import RestInput, UserMessage, CollectingOutputChannel
from sanic.request import Request
from sanic import Sanic, Blueprint, response
from asyncio import Queue, CancelledError
from typing import Text, List, Dict, Any, Optional, Callable, Iterable, Awaitable
from rasa.core import utils
from sanic.response import HTTPResponse

logger = logging.getLogger(__name__)

class BotfrontRestOutput(CollectingOutputChannel):
    @staticmethod
    def _message(
        recipient_id: Text,
        text: Text = None,
        image: Text = None,
        buttons: List[Dict[Text, Any]] = None,
        attachment: Text = None,
        custom: Dict[Text, Any] = None,
    ) -> Dict:
        """Create a message object that will be stored."""

        obj = {
            "recipient_id": recipient_id,
            "text": text,
            "image": image,
            "quick_replies": buttons, # compatibility with Rasa-webchat
            "attachment": attachment,
            "custom": custom,
        }

        # filter out any values that are `None`
        return utils.remove_none_values(obj)

class BotfrontRestInput(RestInput):
    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
        return request.json.get("customData", {})

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
    ) -> Blueprint:
        custom_webhook = Blueprint(
            "custom_webhook_{}".format(type(self).__name__),
            inspect.getmodule(self).__name__,
        )

        # noinspection PyUnusedLocal
        @custom_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @custom_webhook.route("/webhook", methods=["POST"])
        async def receive(request: Request) -> HTTPResponse:
            sender_id = await self._extract_sender(request)
            text = self._extract_message(request)
            should_use_stream = rasa.utils.endpoints.bool_arg(
                request, "stream", default=False
            )
            input_channel = self._extract_input_channel(request)
            metadata = self.get_metadata(request)

            if should_use_stream:
                return response.stream(
                    self.stream_response(
                        on_new_message, text, sender_id, input_channel, metadata
                    ),
                    content_type="text/event-stream",
                )
            else:
                collector = BotfrontRestOutput()
                # noinspection PyBroadException
                try:
                    await on_new_message(
                        UserMessage(
                            text,
                            collector,
                            sender_id,
                            input_channel=input_channel,
                            metadata=metadata,
                        )
                    )
                except CancelledError:
                    logger.error(
                        "Message handling timed out for "
                        "user message '{}'.".format(text)
                    )
                except Exception:
                    logger.exception(
                        "An exception occured while handling "
                        "user message '{}'.".format(text)
                    )
                return response.json(collector.messages)



import logging
import warnings
import uuid
from sanic import Sanic, Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse
from socketio import AsyncServer

from typing import Text, List, Dict, Any, Optional, Callable, Iterable, Awaitable

from rasa.core.channels.channel import UserMessage
from rasa.core.channels.socketio import SocketIOInput, SocketIOOutput, SocketBlueprint

logger = logging.getLogger(__name__)


class WebchatOutput(SocketIOOutput):
    @classmethod
    def name(cls):
        return "webchat"

    async def _send_message(self, socket_id: Text, response: Any) -> None:
        """Sends a message to the recipient using the bot event."""

        await self.sio.emit(self.bot_message_evt, response, room=socket_id)

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        """Send a message through this channel."""

        message_parts = text.split("\n\n")
        for i in range(len(message_parts)):
            text_message = {"text": message_parts[i]}
            if i == len(message_parts) - 1:
                text_message["metadata"] = kwargs.get("metadata", {})
            await self._send_message(self.sid, text_message)

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image to the output"""

        message = {
            "attachment": {"type": "image", "payload": {"src": image}},
            "metadata": kwargs.get("metadata", {}),
        }
        await self._send_message(self.sid, message)

    async def send_text_with_buttons(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        """Sends buttons to the output."""

        message = {
            "text": text,
            "quick_replies": [],
            "quick_replies": buttons,
            "metadata": kwargs.get("metadata", {}),
        }

        await self._send_message(self.sid, message)

    async def send_elements(
        self, recipient_id: Text, elements: Iterable[Dict[Text, Any]], **kwargs: Any
    ) -> None:
        """Sends elements to the output."""

        for element in elements:
            message = {
                "attachment": {
                    "type": "template",
                    "payload": {"template_type": "generic", "elements": element},
                },
                "metadata": kwargs.get("metadata", {}),
            }

            await self._send_message(self.sid, message)

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Sends custom json to the output"""

        json_message.setdefault("room", self.sid)

        await self.sio.emit(
            self.bot_message_evt, **json_message, metadata=kwargs.get("metadata", {})
        )

    async def send_attachment(
        self, recipient_id: Text, attachment: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Sends an attachment to the user."""
        await self._send_message(
            self.sid, {"attachment": attachment, "metadata": kwargs.get("metadata", {})}
        )


class WebchatInput(SocketIOInput):
    @classmethod
    def name(cls):
        return "webchat"

    def __init__(
        self,
        user_message_evt: Text = "user_uttered",
        bot_message_evt: Text = "bot_uttered",
        namespace: Optional[Text] = None,
        session_persistence: bool = False,
        socketio_path: Optional[Text] = "/socket.io",
        cors_allowed_origins="*",
    ):
        self.bot_message_evt = bot_message_evt
        self.session_persistence = session_persistence
        self.user_message_evt = user_message_evt
        self.namespace = namespace
        self.socketio_path = socketio_path
        self.cors_allowed_origins = cors_allowed_origins

    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
        return request.get("customData", {})

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        # Workaround so that socketio works with requests from other origins.
        # https://github.com/miguelgrinberg/python-socketio/issues/205#issuecomment-493769183
        sio = AsyncServer(
            async_mode="sanic", cors_allowed_origins=self.cors_allowed_origins
        )
        socketio_webhook = SocketBlueprint(
            sio, self.socketio_path, "socketio_webhook", __name__
        )

        @socketio_webhook.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @sio.on("connect", namespace=self.namespace)
        async def connect(sid: Text, _) -> None:
            logger.debug(f"User {sid} connected to socketIO endpoint.")

        @sio.on("disconnect", namespace=self.namespace)
        async def disconnect(sid: Text) -> None:
            logger.debug(f"User {sid} disconnected from socketIO endpoint.")

        @sio.on("session_request", namespace=self.namespace)
        async def session_request(sid: Text, data: Optional[Dict]):
            if data is None:
                data = {}
            if "session_id" not in data or data["session_id"] is None:
                data["session_id"] = uuid.uuid4().hex
            await sio.emit("session_confirm", data["session_id"], room=sid)
            logger.debug(f"User {sid} connected to socketIO endpoint.")

        @sio.on(self.user_message_evt, namespace=self.namespace)
        async def handle_message(sid: Text, data: Dict) -> Any:
            output_channel = WebchatOutput(sio, sid, self.bot_message_evt)

            if self.session_persistence:
                if not data.get("session_id"):
                    warnings.warn(
                        "A message without a valid sender_id "
                        "was received. This message will be "
                        "ignored. Make sure to set a proper "
                        "session id using the "
                        "`session_request` socketIO event."
                    )
                    return
                sender_id = data["session_id"]
            else:
                sender_id = sid

            message = UserMessage(
                data["message"],
                output_channel,
                sender_id,
                input_channel=self.name(),
                metadata=self.get_metadata(data),
            )
            await on_new_message(message)

        return socketio_webhook



from rasa_addons.core.actions.action_botfront_mapping import ActionBotfrontMapping
from rasa_addons.core.actions.action_botfront_disambiguation import ActionBotfrontDisambiguation, ActionBotfrontDisambiguationFollowup, ActionBotfrontFallback

from typing import List, Text, Optional, Dict, Any

def actions() -> List["Action"]:
    return [
        ActionBotfrontMapping(),
        ActionBotfrontDisambiguation(),
        ActionBotfrontDisambiguationFollowup(),
        ActionBotfrontFallback()
    ]

actions_bf = {a.name(): a for a in actions()}



from typing import Any

from rasa.nlu.components import Component
from rasa.nlu.training_data import Message

from requests_futures.sessions import FuturesSession


class ActivityLogger(Component):
    name = 'rasa_addons.nlu.components.ActivityLogger'
    defaults = {
        'url': '0.0.0.0',
    }

    def __init__(self, component_config=None):
        super(ActivityLogger, self).__init__(component_config)
        assert 'url' in component_config, 'You must specify the url to use the ActivityLogger component'

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None

        session = FuturesSession()
        params = kwargs.get('request_params', None)
        if params is None:
            return

        if 'model' not in params or params.get('nolog', 'false') not in ['false', '0', '']:
            return

        output = self._message_dict(message)
        output['modelId'] = params.get('model')

        session.post(self.component_config.get('url'), json=output)

    @staticmethod
    def _message_dict(message):
        obj = {
            'text': '',
            'intent': {
                'name': None,
                'confidence': 0,
            },
            'entities': [],
        }
        obj.update(message.as_dict(only_output_properties=True))

        intent = obj['intent']

        obj.update({
            'intent': intent['name'],
            'confidence': intent['confidence'],
        })

        # Botfront expects no intent if intent is None
        if obj['intent'] is None:
            del obj['intent']

        return obj


from typing import Any

from rasa.nlu.components import Component
from rasa.nlu.training_data import Message

from requests_futures.sessions import FuturesSession
import json
import logging

logger = logging.getLogger(__name__)

class HttpLogger(Component):
    name = 'HttpLogger'
    defaults = {
        'url': '0.0.0.0',
        'params': {},  # Params added to the json payload
    }

    def __init__(self, component_config=None):
        super(HttpLogger, self).__init__(component_config)
        assert 'url' in component_config, 'You must specify the url to use the HttpLogger component'
        assert 'model_id' in component_config, 'You must specify the model_id to use the HttpLogger component'

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None

        session = FuturesSession()
        if not message.params: message.params = {}
        if message.params.get('nolog', 'false') in ['true', '1']:
            return

        output = self._message_dict(message)
        for k, v in self.component_config.get('params').items():
            output[k] = v
        output['modelId'] = self.component_config.get('model_id')

        future = session.post(self.component_config.get('url'), json=output)
        response = future.result()
        if response.status_code != 200:
            logger.error('{} Error from API: {}'.format(str(response.status_code), json.loads(response.content)['error']))

    @staticmethod
    def _message_dict(message):
        obj = {
            'text': '',
            'intent': {
                'name': None,
                'confidence': 0,
            },
            'entities': [],
        }
        obj.update(message.as_dict(only_output_properties=True))

        intent = obj['intent']

        obj.update({
            'intent': intent['name'],
            'confidence': intent['confidence'],
        })

        # Botfront expects no intent if intent is None
        if obj['intent'] is None:
            del obj['intent']

        return obj


import logging
from requests_futures.sessions import FuturesSession
logger = logging.getLogger(__name__)


class AskHubAnalytics(object):

    def save(self, latest_turn, tracker, all_events, platform, tracker_store):
        payload = self.get_default_payload(
            tracker,
            all_events[-1]['timestamp'],
            tracker_store.askhub_api_key,
            tracker_store.askhub_response_format,
            latest_turn['language'],
        )

        # User message part
        if latest_turn['user_event']['text']:
            payload['userMessage'] = latest_turn['user_event']['text']
            if 'intent' in latest_turn:
                payload['tags'].append('INTENT_{}'.format(latest_turn['intent'].upper()))
                payload['tags'].append('PARSED')
            else:
                payload['tags'].append('UNPARSED')
            payload['tags'].append('ACTION_QR' if self.is_latest_message_payload(latest_turn) else 'ACTION_NLP')

        # Bot message part
        payload['botResponse'] = latest_turn['bot_events']
        for r in latest_turn['response_names']:
            payload['tags'].append('RESPONSE_{}'.format(r.upper()))
        if platform:
            payload['tags'].append('FILTER_INTERFACE_{}'.format(platform.upper()))
        if 'entities' in latest_turn['parse_data']:
            payload['entities'] = AskHubAnalytics.get_entities(latest_turn)

        AskHubAnalytics.post_async(payload)

    @staticmethod
    def get_entities(latest_turn):
        def convert(entity):
            return {
                'type': 'simple',
                'title': entity['entity'],
                'value': entity['value']
            }

        parse_data = latest_turn['parse_data']
        if 'entities' in parse_data:
            return list(map(convert, parse_data['entities']))
        return None


    @staticmethod
    def response_hook(response, *args, **kwargs):
        # parse the json storing the result on the response object
        if response.status_code != 200:
            reason = response.json() if response.status_code == 400 else response.reason
            logger.warning(
                "Error: post request to Chatbase API failed with status code {}: {}".format(response.status_code,
                                                                                            reason))
            return reason
        else:
            return response.json()

    @staticmethod
    def post_async(payload):

        logger.debug('ASKHUB POST: {}'.format(payload))
        session = FuturesSession()
        session.hooks['response'] = AskHubAnalytics.response_hook
        session.post(' https://api.askhub.io/v1/event', json=payload)

    @staticmethod
    def is_latest_message_payload(latest_turn):
        if latest_turn['user_event']['text']:
            return latest_turn['user_event']['text'].startswith('/')
        return False

    @staticmethod
    def get_default_payload(tracker, timestamp, askhub_api_key, askhub_response_format, language):
        return {
                'token': askhub_api_key,
                'finalUserId': tracker.sender_id,
                'sessionId': tracker.sender_id,
                'timeStamp': round(timestamp * 1000),
                # 'userMessage': event['text'],
                'responseFormat': askhub_response_format,
                'language': language,
                'tags': [],
            }



from typing import Text, List, Dict, Any, Callable, Awaitable, Iterable, Optional
import hashlib
import hmac
import warnings
import logging
from fbmessenger import MessengerClient
from fbmessenger.attachments import Image
from fbmessenger.elements import Text as FBText
from fbmessenger.quick_replies import QuickReplies, QuickReply
from sanic import Blueprint, response
from sanic.request import Request
from typing import Text, List, Dict, Any, Callable, Awaitable, Iterable, Optional

from rasa.core.channels.channel import UserMessage, OutputChannel, InputChannel
from rasa.core.channels.facebook import Messenger, MessengerBot, FacebookInput
from sanic.response import HTTPResponse

logger = logging.getLogger(__name__)

class FBMessengerInput(FacebookInput):

    @classmethod
    def from_credentials(cls, credentials):
        if not credentials:
            cls.raise_missing_credentials_exception()

        return cls(
            credentials.get("verify"),
            credentials.get("secret"),
            credentials.get("page-access-token"),
            credentials.get("fields"),
        )

    def __init__(self, fb_verify, fb_secret, fb_access_token, fields) -> None:
        super().__init__(fb_verify, fb_secret, fb_access_token)
        self.fields = fields

    @staticmethod
    def get_language(user):
        if 'locale' in user:
            split = user['locale'].split('_')
            if len(split) > 0:
                return split[0]
        return None

    def get_metadata(self, request: Request):
        messenger = Messenger(self.fb_access_token, None)
        sender_id = request.json.get('entry')[0].get('messaging')[0].get('sender').get('id')
        user = messenger.client.get_user_data(sender_id, fields=self.fields)
        return {
            'user': user,
            'language': self.get_language(user)
        }




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from rasa.nlu.components import Component
from typing import Any, List, Optional, Text, Dict

from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.model import Metadata
from rasa.nlu.training_data import Message

logger = logging.getLogger(__name__)


class DucklingCrfMerger(Component):
    """Merges Duckling and CRF entities"""

    name = "DucklingCrfMerger"

    provides = []

    defaults = {
        "entities": None,
        "duckling_name": "rasa_addons.nlu.components.duckling_http_extractor.DucklingHTTPExtractor",
    }

    def __init__(self, component_config=None):
        # type: (Text, Optional[List[Text]]) -> None

        super(DucklingCrfMerger, self).__init__(component_config)

    @classmethod
    def create(
        cls, component_config: Dict[Text, Any], config: RasaNLUModelConfig
    ) -> 'DucklingCrfMerger':
        return cls(component_config)

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None
        crf_entities = list(filter(
            lambda e: e["extractor"] == "ner_crf" and e["entity"] in self.component_config["entities"].keys(),
            message.get("entities")))
        indices_to_remove = []

        for index, duck_entity in enumerate(message.get("entities")):
            if duck_entity["extractor"].startswith(self.component_config['duckling_name']):
                # looking for CRF entities surrounding the duckling one matching config settings
                containing_crf = list(filter(
                    lambda e: e["start"] <= duck_entity["start"] and e["end"] >= duck_entity["end"] and duck_entity["entity"] in
                              self.component_config["entities"][e["entity"]], crf_entities))
                # list -> single object
                containing_crf = containing_crf[0] if type(containing_crf) is list and len(containing_crf) > 0 else None
                if containing_crf is not None:
                    # Add duckling value + additional infos
                    containing_crf["value"] = duck_entity["value"]
                    containing_crf["additional_info"] = duck_entity['additional_info']
                    indices_to_remove.append(index)

        # Remove merged duckling entities
        for i in sorted(indices_to_remove, reverse=True):
            del message.get("entities")[i]

    @classmethod
    def load(cls,
             component_meta: Dict[Text, Any],
             model_dir: Text = None,
             model_metadata: Metadata = None,
             cached_component: Optional['DucklingCrfMerger'] = None,
             **kwargs: Any
             ) -> 'DucklingCrfMerger':
        return cls(component_meta)



import json
import logging
import os
from typing import Any, List, Text, Dict
import rasa.utils.io
import re
from rasa.core.actions.action import ACTION_LISTEN_NAME

from rasa.core.domain import Domain
from rasa.core.policies.policy import Policy, confidence_scores_for
from rasa.core.events import SlotSet
from rasa.core.trackers import DialogueStateTracker

logger = logging.getLogger(__name__)

class BotfrontDisambiguationPolicy(Policy):

    @staticmethod
    def _standard_featurizer():
        return None

    def __init__(
        self,
        priority: int = 4,
        disambiguation_trigger: str = "$0 < 2 * $1",
        fallback_trigger: float = 0.30,
        deny_suggestions: Text = "deny_suggestions",
        excluded_intents: List = [], # list of regex e.g ["^chitchat\..*", ...]
        intent_mappings: Dict[Text, Dict[Text, Text]] = {}, # {"basics.yes": {"en": "Yes", "fr": "Oui"}}
        n_suggestions: int = 3,
        disambiguation_title: Dict[Text, Text] = {
            "en": "Sorry, I'm not sure I understood. Did you mean..."
        }
    ) -> None:
        super(BotfrontDisambiguationPolicy, self).__init__(priority=priority)

        self.disambiguation_trigger = disambiguation_trigger
        self.fallback_trigger = fallback_trigger
        self.fallback_default_confidence = 0.30
        self.disambiguation_action = "action_botfront_disambiguation"
        self.disambiguation_followup_action = "action_botfront_disambiguation_followup"
        self.fallback_action = "action_botfront_fallback" # utter_fallback
        self.deny_suggestions = deny_suggestions
        self.excluded_intents = excluded_intents + [self.deny_suggestions]
        self.n_suggestions = n_suggestions
        self.intent_mappings = intent_mappings
        self.disambiguation_title = disambiguation_title

    def train(
        self,
        training_trackers: List[DialogueStateTracker],
        domain: Domain,
        **kwargs: Any
    ) -> None:
        pass

    def generate_disambiguation_message(self, intent_ranking, entities, language):
        intent_names = [intent.get('name', '')
                        for intent in intent_ranking]
        intent_names = [intent for intent in intent_names if len([
                        excl for excl in self.excluded_intents if
                        re.compile(excl).fullmatch(intent) != None]) < 1][:self.n_suggestions]

        mapped_intents = [(name, self.fill_entity(
                            self.intent_mappings.get(name, {language: name})[language],
                            entities
                          )) for name in intent_names]

        entities_json = json.dumps({e.get("entity"): e.get("value") for e in entities}) if len(entities) > 0 else ""

        message_title = self.disambiguation_title[language]
        deny_text = self.intent_mappings.get(self.deny_suggestions, {"en": "Something else"})[language]

        buttons = []
        for intent in mapped_intents:
            buttons.append({'title': intent[1],
                            'type': 'postback',
                            'payload': '/{}{}'.format(intent[0], entities_json)})

        buttons.append({'title': deny_text,
                        'type': 'postback',
                        'payload': '/{}'.format(self.deny_suggestions)})
        return {
            "title": message_title,
            "buttons": buttons
        }
    
    @staticmethod
    def fill_entity(template, entities):
        title = template
        for e in entities:
            title = title.replace('{' + e.get("entity") + '}', e.get("value"))
        title = re.sub(r'{.*?}', r'', title)
        return title

    @staticmethod
    def set_slot(tracker, message):
        if len(message["buttons"]) < 2: return None # abort if only deny_suggestions button would be shown
        try:
            tracker.update(SlotSet("disambiguation_message", value=message))
            result = message
        except Exception as e:
            logger.error("Could not set message slot: {}".format(e))
            result = None
        return result

    @staticmethod
    def _should_disambiguate(intent_ranking, trigger):
        import re
        # pattern to match $0, $1, $2, ... and returning 0, 1, 2,... in match groups
        pattern = re.compile(r"\$(\d)")
        eval_string = trigger
        # matches: an array of intents indices to consider in intent_ranking
        matches = re.findall(pattern, trigger)
        for i in matches:
            # if not enough intents in ranking to apply the rule, policy rule can't be triggered
            if int(i) >= len(intent_ranking): return False
            eval_string = re.sub(r'\$' + i, str(intent_ranking[int(i)]["confidence"]), eval_string)

        return eval(eval_string, {'__builtins__': {}})
    
    @staticmethod
    def _should_fallback(intent_ranking, trigger):
        bonified_ranking = intent_ranking + [{"confidence": 0}]
        return bonified_ranking[0]["confidence"] < trigger

    def _is_user_input_expected(self, tracker: DialogueStateTracker) -> bool:
        return tracker.latest_action_name in [
            self.fallback_action,
            self.disambiguation_action
        ]
    
    def _have_options_been_suggested(self, tracker: DialogueStateTracker) -> bool:
        return tracker.last_executed_action_has(self.disambiguation_action)

    def predict_action_probabilities(
        self, tracker: DialogueStateTracker, domain: Domain
    ) -> List[float]:

        parse_data = tracker.latest_message.parse_data
        metadata = tracker.latest_message.metadata
        entities = parse_data.get("entities", [])
        language = metadata.get("language", "en")
        intent_ranking = parse_data.get('intent_ranking', [])
        can_apply = tracker.latest_action_name == ACTION_LISTEN_NAME
        should_fallback = can_apply and self._should_fallback(intent_ranking, self.fallback_trigger)
        should_disambiguate = can_apply and self._should_disambiguate(
            intent_ranking, self.disambiguation_trigger
        )

        if self._is_user_input_expected(tracker):
            # Shut up and listen
            result = confidence_scores_for(ACTION_LISTEN_NAME, 1.0, domain)

        elif should_fallback:
            logger.debug("Triggering fallback")
            result = confidence_scores_for(self.fallback_action, 1.0, domain)
        
        elif self._have_options_been_suggested(tracker):
            if not should_disambiguate:
                logger.debug("Successfully disambiguated")
                result = confidence_scores_for(self.disambiguation_followup_action, 1.0, domain)
            else:
                logger.debug("Will not disambiguate a second time so fast -- triggering fallback")
                result = confidence_scores_for(self.fallback_action, 1.0, domain)

        elif should_disambiguate:
            logger.debug("Triggering disambiguation")
            disambiguation_message = self.generate_disambiguation_message(intent_ranking, entities, language)
            slot_set = self.set_slot(tracker, disambiguation_message)
            if slot_set:
                result = confidence_scores_for(self.disambiguation_action, 1.0, domain)
            else:
                result = confidence_scores_for(self.fallback_action, 1.0, domain)

        else:
            # Nothing to see here; setting fallback to default confidence
            result = confidence_scores_for(self.fallback_action, self.fallback_default_confidence, domain)

        return result

    def persist(self, path: Text) -> None:
        """Persists the policy to storage."""

        config_file = os.path.join(path, "botfront_disambiguation_policy.json")
        meta = {
            "priority": self.priority,
            "disambiguation_trigger": self.disambiguation_trigger,
            "fallback_trigger": self.fallback_trigger,
            "deny_suggestions": self.deny_suggestions,
            "excluded_intents": self.excluded_intents,
            "n_suggestions": self.n_suggestions,
            "intent_mappings": self.intent_mappings,
            "disambiguation_title": self.disambiguation_title
        }
        rasa.utils.io.create_directory_for_file(config_file)
        rasa.utils.io.dump_obj_as_json_to_file(config_file, meta)

    @classmethod
    def load(cls, path: Text) -> "BotfrontDisambiguationPolicy":
        meta = {}
        if os.path.exists(path):
            meta_path = os.path.join(path, "botfront_disambiguation_policy.json")
            if os.path.isfile(meta_path):
                meta = json.loads(rasa.utils.io.read_file(meta_path))

        return cls(**meta)



import copy
import logging
from collections import defaultdict

from rasa.core.trackers import DialogueStateTracker
from typing import Text, Any, Dict, Optional, List

from rasa.core.nlg.generator import NaturalLanguageGenerator
from rasa.core.nlg.interpolator import interpolate_text, interpolate

logger = logging.getLogger(__name__)


class BotfrontTemplatedNaturalLanguageGenerator(NaturalLanguageGenerator):
    def __init__(self, **kwargs) -> None:
        domain = kwargs.get("domain")
        self.templates = domain.templates if domain else []

    def _templates_for_utter_action(self, utter_action, output_channel, **kwargs):
        """Return array of templates that fit the channel and action."""

        channel_templates = []
        default_templates = []

        for template in self.templates[utter_action]:
            if template.get("language") != kwargs.get("language"):
                continue
            if template.get("channel") == output_channel:
                channel_templates.append(template)
            elif not template.get("channel"):
                default_templates.append(template)

        # always prefer channel specific templates over default ones
        if channel_templates:
            return channel_templates
        return default_templates

    # noinspection PyUnusedLocal
    def _random_template_for(
        self, utter_action: Text, output_channel: Text, **kwargs: Any
    ) -> Optional[Dict[Text, Any]]:
        """Select random template for the utter action from available ones.
        If channel-specific templates for the current output channel are given,
        only choose from channel-specific ones.
        """
        import numpy as np

        if utter_action in self.templates:
            for language in [kwargs.get("language"), kwargs.get("fallback_language")]:
                suitable_templates = self._templates_for_utter_action(
                    utter_action, output_channel, language=language
                )

                if suitable_templates:
                    template = np.random.choice(suitable_templates)
                    return template
            return None
        else:
            return None

    async def generate(
        self,
        template_name: Text,
        tracker: DialogueStateTracker,
        output_channel: Text,
        **kwargs: Any,
    ) -> Optional[Dict[Text, Any]]:
        """Generate a response for the requested template."""

        filled_slots = tracker.current_slot_values()

        fallback_language_slot = tracker.slots.get("fallback_language")
        fallback_language = fallback_language_slot.initial_value if fallback_language_slot else None
        language = tracker.latest_message.metadata.get("language") or fallback_language

        return self.generate_from_slots(
            template_name,
            filled_slots,
            output_channel,
            **kwargs,
            language=language,
            fallback_language=fallback_language,
        )

    def generate_from_slots(
        self,
        template_name: Text,
        filled_slots: Dict[Text, Any],
        output_channel: Text,
        **kwargs: Any,
    ) -> Optional[Dict[Text, Any]]:
        """Generate a response for the requested template."""

        # Fetching a random template for the passed template name
        r = copy.deepcopy(
            self._random_template_for(template_name, output_channel, **kwargs)
        )
        # Filling the slots in the template and returning the template
        if r is not None:
            return self._fill_template(r, filled_slots, **kwargs)
        else:
            return {"text": template_name}

    def _fill_template(
        self,
        template: Dict[Text, Any],
        filled_slots: Optional[Dict[Text, Any]] = None,
        **kwargs: Any,
    ) -> Dict[Text, Any]:
        """"Combine slot values and key word arguments to fill templates."""

        # Getting the slot values in the template variables
        template_vars = self._template_variables(filled_slots, kwargs)

        keys_to_interpolate = [
            "text",
            "image",
            "custom",
            "button",
            "attachment",
            "quick_replies",
        ]
        if template_vars:
            for key in keys_to_interpolate:
                if key in template:
                    template[key] = interpolate(template[key], template_vars)
        return template

    @staticmethod
    def _template_variables(
        filled_slots: Dict[Text, Any], kwargs: Dict[Text, Any]
    ) -> Dict[Text, Any]:
        """Combine slot values and key word arguments to fill templates."""

        if filled_slots is None:
            filled_slots = {}

        # Copying the filled slots in the template variables.
        template_vars = filled_slots.copy()
        template_vars.update(kwargs)
        return template_vars



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import os

import requests
import simplejson
from rasa.nlu.components import Component
from typing import Any, List, Optional, Text, Dict

from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.model import Metadata
from rasa.nlu.training_data import Message

logger = logging.getLogger(__name__)


class EntitiesFilter(Component):
    """Filter entities wrt intent"""

    name = "EntitiesFilter"
    provides = ["entities"]
    defaults = {
        "entities": {}
    }

    def __init__(self, component_config=None):
        # type: (Text, Optional[List[Text]]) -> None

        super(EntitiesFilter, self).__init__(component_config)

    @classmethod
    def create(
        cls, component_config: Dict[Text, Any], config: RasaNLUModelConfig
    ) -> 'EntitiesFilter':
        return cls(component_config)

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None

        # get intent
        intent = message.get("intent")
        if intent is None:
            logger.warn("No intent found")

        # get crf and duckling entities
        message_entities = message.get("entities")
        entities_to_filter = filter(lambda e: e["extractor"] in ["ner_crf", "ner_duckling_http"], message_entities)
        indices_to_remove = []
        for index, entity in enumerate(entities_to_filter):
            if intent["name"] in self.component_config["entities"].keys() and entity["entity"] not in self.component_config["entities"][intent["name"]]:
                indices_to_remove.append(index)

        for i in sorted(indices_to_remove, reverse=True):
            del message.get("entities")[i]

    @classmethod
    def load(cls,
             component_meta: Dict[Text, Any],
             model_dir: Text = None,
             model_metadata: Metadata = None,
             cached_component: Optional['EntitiesFilter'] = None,
             **kwargs: Any
             ) -> 'EntitiesFilter':
        return cls(component_meta)



import logging
import requests
from requests_futures.sessions import FuturesSession
logger = logging.getLogger(__name__)


class ChatbaseAnalytics(object):

    def save(self, latest_turn, tracker, all_events, platform, tracker_store):
        messages = []
        # User Message
        user_message = ChatbaseAnalytics.get_default_payload(tracker, tracker_store,
                                                             latest_turn['user_event']['timestamp'], platform, 'user')
        user_message['message'] = latest_turn['user_event']['text']
        if 'intent' in latest_turn:
            user_message['intent'] = latest_turn['intent']
        no_response = not len(list(filter(lambda r: r.startswith('utter'), latest_turn['response_names'])))
        if no_response or 'intent' not in latest_turn:
            user_message['not_handled'] = True
        messages.append(user_message)
        # Agent Messages
        for response in latest_turn['bot_events']:
            bot_message = ChatbaseAnalytics.get_default_payload(tracker, tracker_store,
                                                                response['timestamp'], platform, 'agent')
            bot_message['message'] = response['text']
            messages.append(bot_message)
        self.post_async(messages)

    @staticmethod
    def get_default_payload(tracker, tracker_store, timestamp, platform, type):
        payload = {
                'api_key': tracker_store.chatbase_api_key,
                'type': type,
                'user_id': tracker.sender_id,
                'time_stamp': timestamp,
                'platform': platform,
                'version': tracker_store.chatbase_version,
            }

        if type == 'user':
            payload['user_id'] = tracker.sender_id

        return payload

    def post_async(self, messages):
        logger.debug('CHATBASE POST: {}'.format(messages))
        session = FuturesSession()
        session.hooks['response'] = self.response_hook
        session.post('https://chatbase.com/api/messages', json={'messages': messages})

    @staticmethod
    def response_hook(response, *args, **kwargs):
        # parse the json storing the result on the response object
        if response.status_code != 200:
            reason = response.json() if response.status_code == 400 else response.reason
            logger.warning(
                "Error: post request to Chatbase API failed with status code {}: {}".format(response.status_code,
                                                                                            reason))
            return reason
        else:
            return response.json()



import logging
import os
import copy
from typing import Optional, Text, Union, List, Dict

from rasa import data
from rasa.core.domain import Domain, InvalidDomain
from rasa.core.interpreter import RegexInterpreter, NaturalLanguageInterpreter
from rasa.core.training.structures import StoryGraph
from rasa.core.training.dsl import StoryFileReader
from rasa.importers import utils
from rasa.importers.importer import TrainingDataImporter
from rasa.nlu.training_data import TrainingData
from rasa.utils import io as io_utils
from  rasa.core.utils import get_file_hash

logger = logging.getLogger(__name__)


class BotfrontFileImporter(TrainingDataImporter):

    def __init__(
        self,
        config_paths: Optional[Dict[Text, Text]] = None,
        domain_path: Optional[Text] = None,
        training_data_path: Optional[Text] = None,
    ):
        # keep only policies in core_config
        self.core_config = {'policies': io_utils.read_config_file(
            config_paths[list(config_paths.keys())[0]]
        )['policies']}
        self._stories_path = os.path.join(training_data_path, 'stories.md')
        
        # keep all but policies in nlu_config
        self.nlu_config = {}
        for lang in config_paths:
            self.nlu_config[lang] = io_utils.read_config_file(config_paths[lang])
            del self.nlu_config[lang]['policies']
            self.nlu_config[lang]['data'] = 'data_for_' + lang # so rasa.nlu.train.train makes the right get_nlu_data call
            self.nlu_config[lang]['path'] = os.path.join(training_data_path, 'nlu', '{}.md'.format(lang))

        self._domain_path = domain_path

    async def get_core_config(self) -> Dict:
        return self.core_config
    
    async def get_nlu_config(self, languages = True) -> Dict:
        if not isinstance(languages, list):
            languages = self.nlu_config.keys()
        return {lang: self.nlu_config[lang] if lang in languages else False for lang in self.nlu_config.keys()}

    async def get_stories(
        self,
        interpreter: "NaturalLanguageInterpreter" = RegexInterpreter(),
        template_variables: Optional[Dict] = None,
        use_e2e: bool = False,
        exclusion_percentage: Optional[int] = None,
    ) -> StoryGraph:

        story_steps = await StoryFileReader.read_from_files(
            [self._stories_path],
            await self.get_domain(),
            interpreter,
            template_variables,
            use_e2e,
            exclusion_percentage,
        )
        return StoryGraph(story_steps)

    async def get_stories_hash(self):
        # Use a file hash of stories file to figure out Core fingerprint, instead of
        # storygraph object hash which is unstable
        return get_file_hash(self._stories_path)

    async def get_nlu_data(self, languages = True) -> Dict[Text, TrainingData]:
        if isinstance(languages, str) and languages.startswith('data_for_'):
            lang = languages.replace('data_for_', '')
            return utils.training_data_from_paths([self.nlu_config[lang]['path']], 'xx')
        if not isinstance(languages, list):
            languages = self.nlu_config.keys()
        td = {}
        for lang in languages:
            try:
                td[lang] = utils.training_data_from_paths([self.nlu_config[lang]['path']], 'xx')
            except ValueError as e:
                if str(e).startswith("Unknown data format"):
                    from rasa.nlu.training_data import TrainingData
                    td[lang] = TrainingData()
        return td

    async def get_domain(self) -> Domain:
        domain = Domain.empty()
        try:
            domain = Domain.load(self._domain_path)
            domain.check_missing_templates()
        except InvalidDomain as e:
            logger.warning(
                "Loading domain from '{}' failed. Using empty domain. Error: '{}'".format(
                    self._domain_path, e.message
                )
            )

        return domain




import logging

logger = logging.getLogger(__name__)


def get_latest_parse_data_language(all_events):
    events = reversed(all_events)
    try:
        while True:
            event = next(events)
            if event['event'] == 'user' and 'parse_data' in event and 'language' in event['parse_data']:
                return event['parse_data']['language']

    except StopIteration:
        return None



from setuptools import setup, find_packages
import os
import io

with io.open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
setup(
    name='rasa_addons',
    version='0.9.4',
    author='Botfront',
    description="Rasa Addons - Components for Rasa and Botfront",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=['pyyaml', 'requests', 'requests_futures', 'fuzzy_matcher', 'fbmessenger'],
    packages=find_packages(exclude=["tests"]),
    licence='Apache 2.0',
    url='https://botfront.io',
    author_email='hi@botfront.io'
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import requests
import logging
import json

from typing import Any

from rasa.nlu.components import Component
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.training_data.message import Message

logger = logging.getLogger(__name__)


class Sweeper(Component):
    name = 'Sweeper'

    defaults = {
        'entity_names': []
    }

    def __init__(self, component_config=None):
        # type: (RasaNLUModelConfig) -> None
        super(Sweeper, self).__init__(component_config)

    def process(self, message, **kwargs):
        # type: (Message, **Any) -> None

        message_entities = message.get("entities")
        sweeped_entities = list(filter(lambda e: e["entity"] not in self.component_config["entity_names"], message_entities))
        message.set("entities", sweeped_entities)

